(in-package :om)

(defclass! tonnetz ()
  ((gens :accessor gens :initform tstandardinv :initarg :gens)
   (chords :accessor chords :initform nil :initarg :chords))
  (:icon 3470)
  (:documentation "Represents the diagram of a triadic Tone Network (Tonnetz) with three generators as well as a chord-list (here, a list of lists of pitches) as arguments."))

; The generators available as pre-defined parameters of the Tone Network.

(defparameter *tonnetz-generators* 
  '(("Standard Generators" tstandardinv) 
    ("Tone Lead Generators" ttoneleadinv)    
    ("Chromatic Generators" tchromaticinv)
    ("Sus2, Sus4 Generators" tsus2sus4inv)))

; Functions to implement basic operability of the Tone Network as an OM class.

(defmethod class-has-editor-p ((self tonnetz)) t)
(defmethod get-editor-class ((self tonnetz)) 'tonnetz-editor)
(defmethod get-type-of-ed-box ((self tonnetz)) 'tonnetz-box)

(defclass tonnetz-box (omboxeditcall)
  ((paths :accessor paths :initform '())))

(defmethod get-slot-names ((self tonnetz))
  (values ; names/accessors
          (list "self" "gens" "chords")
          ; default values
          (list nil 'tstandardinv nil)
          ; doc
          (list "Object" "Generators" "Chords")
          ; menus
          (list nil 
                (list (list 1 *tonnetz-generators*))
                nil)))

; The Tone Network's editor.

(defclass tonnetz-editor (EditorView)
  ((main-view :accessor main-view :initform nil :initarg :main-view)
   (controls-view :accessor controls-view :initform nil :initarg :controls-view)
   (current-chord :accessor currentchord :initform 0)
   (auto-play :accessor autoplay :initform nil)
   (cyclic :accessor cyclic :initform nil)
   (played-chord :accessor playedchord :initform nil)))

(defmethod get-box ((self tonnetz-editor))
  "Returns a reference to the editor's box"
  (ref self))

(defclass tonnetz-control (om-view)
  ((chord-text-view :accessor chord-text-view :initform nil :initarg :chord-text-view)
   (columns-edit :accessor columns-edit)
   (lines-edit :accessor lines-edit)))

(defmethod initialize-instance ((self tonnetz) &rest args)
  (call-next-method)
  (if (symbolp (gens self))
      (setf (gens self) (eval (gens self)))))

(defun from-path (path cols lines)
  "Computes the position in the grid from a Tonnetz path"
  (flet ((tmp (v)
             (case v
               (0 (+ 1 (- lines)))
               (1 (+ lines))
               (2 1)
               (3 (- lines 1))
               (4 (- lines))
               (5 -1))))
    (reduce #'+ (mapcar #'tmp path))))

(defun frompath (path cols lines)
  (mapcar (lambda (x) (from-path x cols lines)) path))

(defmethod refresh-main-view ((self tonnetz-editor) cols lines size gens tile-size)
  "Resets the main view."
  (om-remove-subviews self (main-view self))
  (setf (main-view self)
          (om-make-view 'tonnetz-view 
                        :position (om-make-point 20 20)
                        :columns cols
                        :lines lines
                        :size size
                        :gens gens
                        :tile-size tile-size))
    (om-add-subviews self (main-view self))
    (light-notes self)
    (report-modifications self))

(defmethod initialize-instance ((self tonnetz-editor) &rest args)
  "Initializes the tonnetz-editor window."
  (call-next-method)
  (om-set-bg-color self (om-make-color 0.25 0.25 0.25))
  (setf (main-view self) 
        (om-make-view 'tonnetz-view 
                      :position (om-make-point 20 20)
                      :size (om-make-point 600 600)
                      :gens (gens (object self))))
  (setf (controls-view self) 
        (om-make-view 'tonnetz-control
                      :position (om-make-point (- (w self) 100) 20)
                      :size (om-make-point 80 (- (h self) 40))
                      :bg-color :transparent))

  (om-add-subviews (controls-view self)
                   ;Displays the number of columns.
                   (setf (columns-edit (controls-view self))
                         (om-make-dialog-item 'edit-numbox
                                              (om-make-point 0 0)
                                              (om-make-point 40 20)
                                              (format nil " ~D" (columns (main-view self)))
                                              :fg-color *om-white-color*
                                              :value (columns (main-view self))
                                              ; The following "shadowed" lines cause the selector to fail.
                                              ;:di-action #'(lambda (item)
                                              ;               (set-value item (round (parse-integer (om-dialog-item-text item)) 2)))
                                              :afterfun #'(lambda (item)
                                                            (if (cyclic self)
                                                                (refresh-main-view self 
                                                                                   (* 2 (round (parse-integer (om-dialog-item-text item)) 2))
                                                                                   (lines (main-view self))
                                                                                   (om-view-size (main-view self))
                                                                                   (gens (main-view self))
                                                                                   (tile-size (main-view self)))
                                                              (update-subviews self)))))
                   ; Displays the number of lines.
                   (setf (lines-edit (controls-view self))
                         (om-make-dialog-item 'edit-numbox
                                              (om-make-point 0 25)
                                              (om-make-point 40 20)
                                              (format nil " ~D" (lines (main-view self)))
                                              :value (lines (main-view self))
                                              :fg-color *om-white-color*
                                              :afterfun #'(lambda (item)
                                                            (if (cyclic self)
                                                                (refresh-main-view self 
                                                                                   (columns (main-view self)) 
                                                                                   (* 2 (round (parse-integer (om-dialog-item-text item)) 2))
                                                                                   (om-view-size (main-view self))
                                                                                   (gens (main-view self))
                                                                                   (tile-size (main-view self)))
                                                              (update-subviews self)))))
                    ; Indicates the length of the window (number of chords activated.)
                    (setf (chord-text-view (controls-view self))
                          (om-make-dialog-item 'om-text-view
                                               (om-make-point 10 55)
                                               (om-make-point 120 20)
                                               (chord-text self)
                                               :fg-color *om-white-color*
                                               :bg-color :transparent
                                               ))
                    ; Lists the available generators when clicked on, toggles and displays the Tone Network according to the current ones.
                    (om-make-dialog-item 'om-pop-up-dialog-item
                                               (om-make-point 10 85)
                                               (om-make-point 120 20)
                                               ""
                                               :range (mapcar 'car *tonnetz-generators*)
                                               :value NIL ; Q: HOW TO FIND THE VALUE FROM THE TONNETZ?
                                               :di-action #'(lambda (item)
                                                              (setf (gens (object self))
                                                                    (eval (cadr (nth (om-get-selected-item-index item) *tonnetz-generators*))))
                                                              (refresh-main-view self 
                                                                          (columns (main-view self))
                                                                          (lines (main-view self))
                                                                          (om-view-size (main-view self))
                                                                          (gens (object self))
                                                                          (tile-size (main-view self)))
                                                              ))
                    ; Toggles auto-play.
                    (om-make-dialog-item 'om-check-box
                                         (om-make-point 10 165)
                                         (om-make-point 120 40)
                                         "Auto-play"
                                         :fg-color *om-light-gray-color*
                                         :value (autoplay self) 
                                         :di-action #'(lambda (item)
                                                        (setf (autoplay self)
                                                              (om-checked-p item)))
                                         )
                    (om-make-dialog-item 'om-check-box
                                         (om-make-point 10 205)
                                         (om-make-point 120 40)
                                         "Cyclic" ; otherwise, Simile-infinite.
                                         :fg-color *om-light-gray-color*
                                         :value (cyclic self) 
                                         :di-action #'(lambda (item)
                                                        (setf (cyclic self)
                                                              (om-checked-p item)))
                                         )
                    (om-make-dialog-item 'om-text-view
                                         (om-make-point 10 125)
                                         (om-make-point 20 20)
                                         "#"
                                         :fg-color *om-light-gray-color*
                                         :bg-color :transparent
                                         )
                    ; Allows the user to change the number of displayed chords — the trace.
                    (om-make-dialog-item 'edit-numbox
                                   (om-make-point 30 122)
                                   (om-make-point 40 20) 
                                   (format nil " ~D" (max-level (main-view self)))
                                   :bg-color *om-black-color*
                                   :fg-color *om-white-color*
                                   :value (max-level (main-view self))
                                   :min-val 1 :max-val 100
                                   :afterfun #'(lambda (item)
                                                 (setf (max-level (main-view self)) (value item))
                                                 (om-set-dialog-item-text (chord-text-view (controls-view self)) (chord-text self))
                                                 ; Changes every remaining chord's hue, as it applies the same gradient to a different number of chords.
                                                 (erase-notes self)
                                                 (light-notes self)
                                                 (om-invalidate-view (main-view self)))
  ))
  (om-add-subviews self (main-view self) (controls-view self))
  ; Highlights the notes from the chords if there are any.
  (light-notes self)
  (om-invalidate-view (main-view self)))

(defun tiles-num (cols rows)
  "Returns the number of tiles, given the number of columns and lines."
  (- (* cols rows) (floor (/ cols 2))))

(defmethod draw-light ((self tonnetz-editor) poslist)
  "Toggles the state of each tile visible in the main view."
  (loop for k in poslist do
        (let ((i (cadr k))
              (l (car k)))
          (loop for x in l do
                (push i (level (nth x (tiles (main-view self)))))
                (setf (color (nth x (tiles (main-view self)))) (om-make-color
                                                                              (+ 0.3125 (/ (+ 0.25 i) (max-level (main-view self))))
                                                                              (+ 0.3125 (/ (+ 0.25 i) (max-level (main-view self)))) 
                                                                              0.3125))))))

(defmethod light-notes ((self tonnetz-editor))
  "Highlights the selected notes, given that the Tone Network has been correctly initialized."
  (let* ((tiles (tiles (main-view self)))
         (tchs (last (chords (object self)) (max-level (main-view self))))
         (pchs (mapcar #'netz-permute tchs))
         (gs (gens (main-view self)))
         (tpaths (mapcar (lambda (tmp) (mapcar (lambda (c) (list c (netz-computepath gs c))) tmp)) pchs)) ; Paths associated to a given chord
         (chordnpaths (mapcar #'most-compact tpaths))
         (paths (mapcar #'cadr chordnpaths))
         (chs (mapcar #'car chordnpaths))
         (cols (columns (main-view self)))
         (lines (lines (main-view self))))
    ;(setf (max-level (main-view self)) (length paths))
    (if chs
        (progn 
          (draw-light self (print (compute-paths self paths cols lines chs gs)))
      '()))))

(defmethod compute-paths ((self tonnetz-editor) paths cols lines chs gs)
  (let ((index (position 0 (mapcar #'note (tiles (main-view self))) :start (* 6 lines))) ; We have to search for a tile with a 0 note (C). The best would be to look for such a tile, with the requirement that the path fits into the tonnetz.
        (tmpl '())
        (bestlastnote 0)
        (poslist '()))
    (setf (paths (main-view self)) paths)
    (loop for c in paths
          for i in (iota (min (max-level (main-view self)) (length paths))) do
          (if (/= i 0)
              (setf bestlastnote (car (min-ch 
                                       (mapcar (lambda (note) (list note (cadr (distance-note2chord note (nth i chs) gs)))) (nth (- i 1) chs)) 
                                       #'cadr))
                    index (nth (position bestlastnote (nth (- i 1) chs)) tmpl)))
          (let* ((onenote (from-path (compute-path gs (netz-interval bestlastnote (car (nth i chs)))) cols lines))
                 (l (scanl-minus #'+ (cons (setf index (+ onenote index)) (frompath c cols lines)))))
            (push (list l i) poslist)
            (setf tmpl l)
            ))
    poslist))

; Bound to the button action. To be removed, maybe. But necessary for tiles and stuff.

(defmethod generate-chords ((self tonnetz-editor))
  (let ((ton (object self))
        (tiles (tiles (main-view self))))
    (setf (chords ton)
          (remove-if #'null
                     (loop for x in (iota (max-level (main-view self))) collect
                           (loop for n in (remove-if (lambda (l) (not (member x (level l)))) tiles) collect
                                 (note n)))))
    ; The following "shadowed" lines might be wanted or not. And, maybe, even add a boolean?
    ;(erase-notes self)
    ;(light-notes self)
    ; Invalidates the editor and reports the modification to the box.
    (update-panel self t)))

(defmethod erase-notes ((self tonnetz-editor))
  "Erases all notes from the editor's main view."
  (mapcar (lambda (el) 
            (progn
              (setf (level el) '())
              (setf (color el) (om-make-color 1 1 1)))) (tiles (main-view self))))

(defmethod chord-text ((self tonnetz-editor))
  "Displays, in a string, the index of the currently selected chord."
  (concatenate 'string "Chord: " (write-to-string (+ 1 (currentchord self))) "/" (write-to-string (max-level (main-view self)))))

(defmethod netz-transpose-chords ((self tonnetz-editor) n-gen)
  "Transposes the set of chords given a generator (by its index in a list of generators.)"
  (let ((trans (nth n-gen (gens (object self)))))
    (setf (chords (object self))
          (mapcar (lambda (l) (mapcar (lambda (el) (funcall trans el)) l)) (chords (object self)))))
  (erase-notes self)
  (light-notes self))

(defun netz-convert2mc (tchord)
  "Converts a list of pitches (from a Tone Network) into a list of midicents."
  (mapcar (lambda (note) (* 100 (+ (* 12 5) note))) tchord))

(defmethod netz-play-chord ((self tonnetz-editor))
  "Plays the currently selected chord."
  (let* ((onechord (nth (currentchord self) (chords (object self))))
          (chord (netz-convert2mc onechord)))
    (setf c (make-instance 'chord :lmidic chord))
    (play c)
    ))

(defmethod netz-modify-current ((self tonnetz-editor) n)
  "Modifies the currently selected chord and plays it if selected by the user."
  (setf (currentchord self) n)
  (om-set-dialog-item-text (chord-text-view (controls-view self)) (chord-text self))
  (when (autoplay self)
    (netz-play-chord self)))

; Defines the key commands available.

(defmethod handle-key-event ((self tonnetz-editor) key)
  (case key 
    (:om-key-right ; Toggles the next chord, and plays it if "Auto-play" is enabled.
         (netz-modify-current self (mod (+ 1 (currentchord self)) (max-level (main-view self)))))
    (:om-key-left ; Toggles the first chord, and plays it if "Auto-play" is enabled.
         (netz-modify-current self 0))    
    (#\Space ; Plays the current chord.
         (netz-play-chord self))
    (#\0 ; Clears the Tone Network's main view.
         (erase-notes self))
    (#\- ; Decreases the Zoom Level (the tiles' size.)
         (resize-tiles self 0.875))
    (#\+ ; Increases the Zoom Level (the tiles' size.)
         (resize-tiles self 1.125))
    (#\o ; Toggles the opening chord, and plays it if "Auto-play" is enabled — identical to "left-arrow" key command, only here to accompany the "p" key command.
         (netz-modify-current self 0))    
    (#\p ; Plays the current chord and toggles the next — only use it if "Auto-play" is disabled, otherwise use the "space" and "right-arrow" key commands.
         (netz-play-chord self)
         (netz-modify-current self (mod (+ 1 (currentchord self)) (max-level (main-view self)))))
    (#\z (print "Zzzz"))    

    ; Transposition key commands.

    (#\t
     (netz-transpose-chords self 5))
    (#\y
     (netz-transpose-chords self 2))
    (#\h
     (netz-transpose-chords self 3))
    (#\f 
     (netz-transpose-chords self 4))
    (#\v 
     (netz-transpose-chords self 0))
    (#\b
     (netz-transpose-chords self 1))

    (otherwise (print nil)))
  (om-invalidate-view self)
  )

; Redraws a Tone Network with resized tiles by a factor of one's choice. 

(defmethod resize-tiles ((self tonnetz-editor) factor)
  "Resizes the tiles by a given factor."
  (let ((tsize (* factor (tile-size (main-view self))))
        (viewsize (om-view-size (main-view self))))
    (om-remove-subviews self (main-view self))
    (setf (main-view self)
          (om-make-view 'tonnetz-view :position (om-make-point 20 20)
                        :size viewsize
                        :gens (gens (object self))
                        :tile-size tsize))
    (om-add-subviews self (main-view self))
    (light-notes self)))

(defmethod update-editor-after-eval ((self tonnetz-editor) val)
  ; After evaluation, completely redraws the Tone Network.
  (call-next-method)
  (erase-notes self)
  (light-notes self))

(defmethod update-subviews ((self tonnetz-editor))
  ; Called when resizing.
  (call-next-method)
  (om-set-view-size (main-view self) (om-subtract-points (om-view-size self) (om-make-point 200 40)))
  (om-set-view-position (controls-view self) (om-make-point (- (w self) 160) 20))
  (om-set-view-size (controls-view self) (om-make-point 140 (- (h self) 40)))
  ; Redraws the main view.
  (refresh-main-view self 
                     (columns (main-view self))
                     (lines (main-view self))
                     (om-view-size (main-view self))
                     (gens (object self))
                     (max (/ (- (w self) 160) (columns (main-view self))) (/ (h self) (lines (main-view self))))))

; Functions to draw the box.

(defun gen-to-vect (g)
  "Associates a direction (a two-dimmensional vector) to each generator."
  (let ((x 35)
        (y 40))
    (apply #'om-make-point
           (case g
             (0 (list (- x) (/ y -2)))
             (1 (list x (/ y -2)))
             (2 (list 0 (- y)))
             (3 (list x (/ y 2)))
             (4 (list (- x) (/ y 2)))
             (5 (list 0 y))))))

(defmethod draw-obj-in-rect ((self tonnetz) x1 x2 y1 y2 params view)
  ; Draws the hexagons.
  (let ((editor (editorframe (object (om-view-container view)))))
    (if editor
        (let* ((i (currentchord editor))
               (stiles (remove-if (lambda (ti) (not (find i (level ti)))) (tiles (main-view editor))))
               (chord (mapcar #'note stiles))
               (pchord (netz-permute chord))
               (paths (mapcar (lambda (ch) (netz-computepath (gens (object editor)) ch)) pchord))
               (path (mostcompact paths))
               (tcoords (mapcar (lambda (ti) (if (listp ti)
                                                 (reduce #'om-add-points (mapcar #'gen-to-vect ti))
                                               (gen-to-vect ti))) path))
               (coords (scanl #'om-add-points (om-make-point (/ x2 2) (/ y2 2)) tcoords)))
          (print coords)
          (om-with-focused-view view
            (loop for point in coords do
                  (let ((xc (om-point-x point))
                        (yc (om-point-y point)))
                    (om-draw-ellipse xc yc 19 19)
                    (om-with-fg-color view (om-make-color 0.875 0.875 0)
                      (om-fill-ellipse xc yc 19 19)
                      ))))))))


; Toolbox.

; Converts a Chord-Seq into a Tone Network.

(defmethod! cseq2netz ((cs chord-seq))
  :doc "Converts a Chord-Seq into a Tone Network."
  :indoc '("Chord-Seq")
  :outdoc '("The Tone Network representing the Chord-Seq's chords.")
  :icon 3470
  (make-instance 'tonnetz :chords (mapcar (lambda (l) (mapcar (lambda (el) (mod (/ el 100) 12)) l)) (lmidic cs))))

; Converts a Tone Network into a Chord-Seq.

(defmethod! netz2cseq ((ton tonnetz))
  :doc "Converts a Tone Network into a Chord-Seq."
  :indoc '("Tone Network")
  :outdoc '("The Chord-Seq representing the Tone Network's chords.")
  :icon 3470
  (make-instance 'chord-seq :lmidic (mapcar (lambda (chr) (mapcar (lambda (note) (* 100 (+ (* 12 5) note))) chr)) (chords ton))))

; Transposes a Tone Network given a direction defined by one of its generators.

(defmethod! tonnetz-transpose ((ton tonnetz) n-gen)
   :doc "Transposes a Tone Network's path following the axis defined by one of its generators."
   :indoc '("Tone Network" "Integer : index of one of its generators.")
   :outdoc '("The transposed Tone Network.")
   :icon 3470
   :initvals '(nil 0)
   :menuins '((1 (("1" 0) ("2" 1) ("3" 2) ("4" 3) ("5" 4) ("6" 5))))
   (make-instance 'tonnetz :gens (gens ton) :chords (tonnetz-transpose-chords (chords ton) (gens ton) n-gen)))

; Creates a list of generators. (Not Working!)

(defun make-generators (n1 n2 n3)
  "Creates a list of generators given three modulo 12 natural numbers (representing intervals)."
  (list (netz-interval 0 n1) (netz-interval 0 n2) (netz-interval 0 n3)))

; Tiles

; Classes and functions implementing the tiles and the drawing thereof.

; Represents a tile (note) in a Tone Network.

(defclass note-view (om-item-view)
  ((note :accessor note :initform 0 :initarg :note)
   (column :accessor col :initform 0 :initarg :column)
   (row :accessor row :initform 0 :initarg :row)
   (level :accessor level :initform '())
   (color :accessor color :initform (om-make-color 1 1 1))))

; Draws a tile.

(defmethod om-draw-contents ((self note-view))
  (let* ((x3 (w self))
         (x1 (/ x3 3))
         (c 4)
         (ystart (+ c (/ (h self) 6)))
         (y1 (- (h self) ystart))
         (y2 (/ (+ y1 ystart) 2))
         (name (netz-p2n (note self))))
    (om-with-focused-view self
      (om-with-fg-color
          self (if (member (currentchord (om-view-container (om-view-container self))) (level self)) (om-make-color 1 1 0) (color self)) 
        (om-fill-ellipse (/ (w self) 2) (/ (h self) 2) (- (/ (w self) 2) 2) (- (/ (w self) 2) 2)))
      (om-with-fg-color
          self (if (member (currentchord (om-view-container (om-view-container self))) (level self)) (om-make-color 1 1 1) (om-make-color 0.25 0.25 0.25))
        (om-draw-ellipse (/ (w self) 2) (/ (h self) 2) (- (/ (w self) 2) 2) (- (/ (w self) 2) 2))
        )
      ; Defines a cell's font, and its location within the cell.
     (om-with-font (om-make-font "Arial" (/ (w self) 3))
                   (case (length name)
                     (1 (om-draw-string (+ 3 x1) (+ 4 y2) name))
                     (2 (om-draw-string (+ 0 x1) (+ 4 y2) name)))))))

(defmethod add-to-currentchord ((self note-view))
  "Adds a note to the currently selected chord and displays it"
  (let ((cchord (currentchord (om-view-container (om-view-container self)))))
    (if (member cchord (level self))
        (setf (level self) (remove cchord (level self)))
      (push cchord (level self))))
  (setf (color self)
        (if (null (level self))
            (om-make-color 1 1 1)
          (om-make-color 
                         (+ 0.3125 (/ (+ 0.25 (apply #'max (level self))) (max-level (main-view (om-view-container (om-view-container self)))))) 
                         (+ 0.3125 (/ (+ 0.25 (apply #'max (level self))) (max-level (main-view (om-view-container (om-view-container self)))))) 
                         0.3125)))
  (generate-chords (om-view-container (om-view-container self)))
  (om-invalidate-view self))

(defmethod create-chord ((self note-view))
  "Creates a new chord and refreshes the main view"
  (let ((ed (om-view-container (om-view-container self))))
    (setf (chords (object (om-view-container (om-view-container self))))
          (append (chords (object (om-view-container (om-view-container self))))
                  (list (list (note self)))))
    (setf (currentchord ed) (max-level (main-view ed)))
    (refresh-main-view ed
                       (columns (main-view ed))
                       (lines (main-view ed))
                       (om-view-size (main-view ed))
                       (gens (object ed))
                       (tile-size (main-view ed)))
  ))

; Changes the state of a tile
(defmethod om-view-click-handler ((self note-view) position)
  ; Is the ctrl key pressed or not?
  (if (om-command-key-p)
      (create-chord self)
    (add-to-currentchord self)))

; View

; Classes and functions implementing the drawing of the Tone Network.

; The main view.

(defclass tonnetz-view (om-view)
  ((tiles :accessor tiles :initform nil)
   (paths :accessor paths :initform nil)
   (gens :accessor gens :initarg :gens)
   (notes :accessor notes :initform nil)
   (columns :accessor columns :initform 36 :initarg :columns)
   (lines :accessor lines :initform 24 :initarg :lines)
   (max-level :accessor max-level :initform 4)
   (tile-size :accessor tile-size :initform 36 :initarg :tile-size)))

 ; Draws the tonnetz

(defmethod initialize-instance ((self tonnetz-view) &rest args)
  (call-next-method)
  (om-set-bg-color self (om-make-color 0.75 0.75 0.75))
  (add-tiles self)
  (setf (tiles self) (reverse (tiles self)))
    ;(apply #'om-add-subviews (cons self (tiles self))) ; Too long list
  (mapcar (lambda (el) (om-add-subviews self el)) (tiles self)))

(defmethod add-tiles ((self tonnetz-view))
  "Adds the tiles to the view"
  (let* ((cols (columns self))
         (lines (lines self))
         (width (tile-size self))
         (height (* 1.5 width))
         (shift (/ height 2)))
    ; Initializing the tiles
    (loop for x in (iota cols)
          for n1 in (netz-computemelody (gens self) 0 (mapcar #'list (apply #'append (make-list cols :initial-element '(1 3)))))
          do
          (loop for y in (iota (- lines (if (oddp x) 1 0)))
                for n2 in (netz-computemelody (gens self) n1 (mapcar #'list (make-list cols :initial-element 2)))
                do
                (setf (tiles self) (cons (om-make-view 'note-view
                                                       :position (om-make-point (* x width) (+ (if (oddp x) shift 0) (* y height)))
                                                       :column x
                                                       :row y
                                                       :size (om-make-point width height)
                                                       :bg-color :transparent
                                                       :note n2)
                                         (tiles self)))))))

(defmethod om-draw-contents ((self tonnetz-view))
  (let* ((width (tile-size self))
         (height (* 1.5 (tile-size self)))
         ; Diagonals (from 0,0 to x,y)
         (startx1 0)
         (starty1 (+ 4 (/ height 6)))
         (endx1 (* width (columns self)))
         (endy1 (+ starty1 (* height (/ (columns self) 2))))
         ; Diagonals (from 0,y to x,0)
         (startx2 0)
         (starty2 (+ starty1 (* height (/ (columns self) 2))))
         (endx2 (* width (columns self)))
         (endy2 (+ 4 (/ height 6)))
         )
  (om-with-focused-view self
    (om-with-fg-color self (om-make-color 0.5 0.5 0.5)
      ; Diagonals (from 0,0 to x,y)
      (loop for x in (mapcar (lambda (e) (* 2 e)) (iota (/ (columns self) 2))) do
            (om-draw-line (+ startx1 (* x width)) starty1 (+ (* x width) endx1) endy1))
      (loop for y in (iota (lines self)) do
            (om-draw-line startx1 (+ (* y height) starty1) endx1 (+ (* y height) endy1)))
      ; Diagonals (from 0,y to x,0)
      (loop for x in (mapcar (lambda (e) (- (* 2 e) 1)) (iota (/ (columns self) 2))) do
            (om-draw-line (- startx2 (* x width)) starty2 (- endx2 (* x width)) endy2))
      (loop for y in (iota (lines self)) do
            (om-draw-line startx2 (+ (* y height) (+ (/ height 2) starty2)) endx2 (+ (* y height) (+ (/ height 2) endy2))))
      ; Vertical lines
      (loop for x in (iota (columns self)) do
            (om-draw-line (+ (/ width 2) (* x width)) (+ (if (evenp x) (/ height 2) 0) starty1) (+ (/ width 2) (* x width)) (+ (if (evenp x) (/ height 2) 0) starty1 (* height (lines self)))))
                ))))


; Utilities

; Defines a few functions which make us able to work on Tonnetze
; Transpositions are here represented by functions rather than numbers,
; so that we keep the algebraic group structure. These functions act on
; the set of notes, which are simply numbers (from 0 to 11).
; paths in the Tonnetz are lists of positive integers, representing the
; generators (to model a path, we take the position of the generator in 
; the list of generators, and use it).

(defun netz-add-mod12 (a b)
;  "Modulo 12 addition, represents the addition in Z/12Z"
  (mod (+ a b) 12))

(defun iota (n &optional (startpoint 0))
;  "Builds a list of n elements beginning at startpoint."
  (labels ((tmpi (n sa acc)
             (if (<= n 0) acc (funcall #'tmpi (- n 1) (+ sa 1) (append acc (list sa))))))
    (funcall #'tmpi n startpoint nil)
    ))

(defun netz-transpose (a)
;  "Creates a function which transposes a given pitch."
  (lambda (b) (mod (+ a b) 12)))

(defun netz-inverse (i)
;  "Creates a function of the inverse of a given interval."
  (netz-transpose (- 12 (funcall i 0))))

; The null element of the transposition group.

(defvar el-null (netz-transpose 0))

; Generators of the standard riemannian Tone Network (Tonnetz): the minor and major thirds, and the perfect fifth.

(defvar tstandard (mapcar #'netz-transpose '(3 4 7)))
(defvar tstandardinv (append tstandard (mapcar #'netz-inverse tstandard)))

; Generators of non-riemannian, triadic Tone Networks (Tonnetze).

(defvar ttonelead (mapcar #'netz-transpose '(1 2 3)))
(defvar ttoneleadinv (append ttonelead (mapcar #'netz-inverse ttonelead)))
(defvar tchromatic (mapcar #'netz-transpose '(1 1 2)))
(defvar tchromaticinv (append tchromatic (mapcar #'netz-inverse tchromatic)))
(defvar tsus2sus4 (mapcar #'netz-transpose '(2 5 7)))
(defvar tsus2sus4inv (append tsus2sus4 (mapcar #'netz-inverse tsus2sus4)))

(defun netz-compose (a b)
;  "Represents the operation of the transposition group."
  (lambda (x) (funcall a (funcall b x))))

(defun netz-interval (a b)
;  "Computes a transposition of the distance between two given pitches."
  (netz-transpose (- b a)))

(defun netz-equal (f1 f2)
;  "Computes a boolean of the predicate that both given transpositions yield the same values, i.e. are equal."
  (equal (funcall f1 0) (funcall f2 0)))

(defun netz-traverse (is ins)
;  "Computes a transposition, given the generators and a path."
  (reduce #'netz-compose (mapcar (lambda (n) (nth n is)) ins) :initial-value el-null))

(defun tmpcompute (is i p)
  (let* ((ni (length is))
         (v (apply #'append (mapcar (lambda (x) (mapcar (lambda (b) (cons x b)) p)) (iota ni))))
         (l (remove-if (lambda (y) (not (netz-equal i (netz-traverse is y)))) v)))
    (if (netz-equal i (netz-transpose 0))
        '()
        (if (null l)
            (tmpcompute is i v)
          l))))

(defun netz-pairs (l)
  (if (or (null (cdr l)) (null l))
      '()
    (cons (list (car l) (cadr l)) (netz-pairs (cdr l)))))

(defun compute-path (is i)
    (reverse (car (tmpcompute is i (list '())))))

(defun netz-computepath (is ps)
;  "Computes a path in a Tone Network, given the generators and a melody."
  (let ((ins (mapcar (lambda (l) (netz-interval (car l) (cadr l))) (netz-pairs ps))))
    (mapcar (lambda (x) (compute-path is x)) ins)))

(defun scanl (f q ls)
;  "Similar to a reduce, but successively folds over lists of increasing length."
  (if (null ls)
      (list q)
    (cons q (scanl f (funcall f q (car ls)) (cdr ls)))))

(defun scanl-minus (f ls)
;  "scanl without the start argument."
  (if (null ls)
      '()
    (scanl f (car ls) (cdr ls))))

(defun netz-computemelody (gens start paths)
;  "Computes a melody, given the generators, the starting note, and a list of paths."
   (cons start (mapcar (lambda (i) (funcall i start)) (scanl-minus #'netz-compose (mapcar (lambda (path) (netz-traverse gens path)) paths)))))

(defun netz-p2n (n)
;  "Maps a pitch onto a note — important for the UI."
  (case n
    (0 "C")
    (1 "C#")
    (2 "D")
    (3 "D#")
    (4 "E")
    (5 "F")
    (6 "F#")
    (7 "G")
    (8 "G#")
    (9 "A")
    (10 "A#")
    (11 "B")))

; Tonnetz Transformations

; Transposition

(defun tonnetz-transpose-chords (chords gens n-gen)
;  "Transposes a list of chords along an axis, given by the index of the generator."
  (mapcar (lambda (l) (mapcar (lambda (el) (funcall (nth n-gen gens) el)) l)) chords))

(defun tonnetz-rotate-pos (n-gen)
  (case n-gen
    (0 1)
    (1 3)
    (2 5)
    (3 2)
    (4 0)
    (5 4)))

(defun tonnetz-rotate-pos-inverse (n-gen)
  (case n-gen
    (1 0)
    (3 1)
    (5 2)
    (2 3)
    (0 4)
    (4 5)))

; Not finished.
(defun tonnetz-rotate-chords (chords gens angle)
  "Rotation in a Tonnetz, angle is -1 or +1."
  (let* ((frotate (if (< angle 0) #'tonnetz-rotate-pos-inverse #'tonnetz-rotate-pos))
         (chs-gens (funcall #'frotate (rmapcar (lambda (i) (position i gens :test netz-equal)) (mapcar (lambda (chr) (netz-computepath gens chr)) chords))))
         (ins (mapcar (lambda (p) (netz-interval (car (last (car p))) (cadr p))) (netz-pairs chords))) ; Distances between chords
         (fstchr (mapcar (lambda (note) (funcall (nth note gens) (caar chords))) chs-gens))
         (result (list fstchr)))
    (loop for c in (cdr chs-gens)
          for i in ins
          
          )
  result
  ))

(defun netz-permute (list)
;  "Permutation function from Rosetta code. Returns all the permutations of a given list."
  (if list
    (mapcan #'(lambda (x)
    (mapcar #'(lambda (y) (cons x y))
      (netz-permute (remove x list))))
      list)
    '(())))

(defun min-ch (box access)
  (labels ((tmpmin-ch (box acc)
             (if (null box)
                 acc
               (tmpmin-ch (cdr box) (if (< (funcall access (car box)) (funcall access acc))
                                                  (car box)
                                                acc)))))
    (tmpmin-ch box (car box))))
      
(defun mostcompact (box)
;  "Selects the most compact chord in a list of chords. The algorithm selects the list for which the sum of the length of the sublists is lower."
  (car (min-ch
        (mapcar (lambda (el) (list el (reduce #'+ (mapcar (lambda (e) (if (listp e) (- (length e) 1) 0)) el)))) box)
        #'cadr)
       ))

(defun most-compact (box)
  (car (min-ch
        (mapcar (lambda (el) (list el (reduce #'+ (mapcar (lambda (e) (if (listp e) (- (length e) 1) 0)) (cadr el))))) box)
        #'cadr)))

(defun distance-note2chord (note chord gens)
;  "Returns the distance between a chord and a note, and the note of the chord for which the distance is minimal"
  (min-ch 
   (mapcar (lambda (el)
             (list (car el) (length (cadr el))))
           (mapcar (lambda (nt) (list nt (if (= note nt) '() (car (netz-computepath gens (list nt note)))))) chord))
   #'cadr))

; THE END... For now.